/* --------------------------------------------------------------------- 
 * Implementation of class Valkyrie                  valkyrie_object.cpp
 * Valkyrie-specific options / flags / fns
 * ---------------------------------------------------------------------
 * This file is part of Valkyrie, a front-end for Valgrind
 * Copyright (c) 2000-2005, OpenWorks LLP <info@open-works.co.uk>
 * This program is released under the terms of the GNU GPL v.2
 * See the file LICENSE.GPL for the full license details.
 */

#include "valkyrie_object.h"
#include "valgrind_object.h"
#include "tool_object.h"
#include "xml_parser.h"        // xmlFormatCheck

#include "vk_config.h"
#include "vk_messages.h"
#include "vk_utils.h"          // vk_assert(), vk_strcmp(), vkPrint()
#include "html_urls.h"
#include "vk_include.h"        // VK_VERSION

#include <qapplication.h>


/* class Valkyrie ------------------------------------------------------ 
   The key value of HELP_OPT is used:
   - to exclude the option from configEntries;
   - to determine when a shortFlag is needed for popt.
   The key value of NOT_POPT is used:
   - to exclude the option from popt options.  
*/
Valkyrie:: ~Valkyrie() { }


Valkyrie::Valkyrie()
  : VkObject( "Valkyrie", "Valkyrie", Qt::Key_unknown, false ) 
{
  /* init vars */
  runMode = modeNotSet;
  valgrind   = 0;


  addOpt( HELP_OPT,    Option::ARG_NONE,   Option::NONE,
          "",          'h',                "help", 
          "",          "",                 "", 
          "",          "show this help message and exit", urlNone );
  addOpt( HELP_OPT,    Option::ARG_NONE,   Option::NONE,
          "",          'v',                "version", 
          "",          "",                 VK_VERSION, 
          "",          "display version information and exit", urlNone );
  addOpt( HELP_OPT,    Option::ARG_NONE,   Option::NONE,
          "",          'V',                "valgrind-opts", 
          "",          "",                 "", 
          "",          "Show valgrind options too, and exit", urlNone );
  addOpt( TOOLTIP,     Option::NOT_POPT,   Option::CHECK, 
          "valkyrie",  '\0',               "show-tooltips", 
          "",          "true|false",       "true", 
          "Show tooltips",      "",        urlValkyrie::toolTips );
  addOpt( PALETTE,     Option::NOT_POPT,   Option::CHECK, 
          "valkyrie",  '\0',               "use-vk-palette", 
          "",          "true|false",       "true", 
          "Use valkyrie's palette",   "",  urlValkyrie::palette );
  addOpt( ICONTXT,     Option::NOT_POPT,   Option::CHECK, 
          "valkyrie",  '\0',               "show-butt-text", 
          "",          "true|false",       "true", 
          "Show toolbar text labels",  "", urlValkyrie::toolLabels );
  addOpt( FONT_SYSTEM,  Option::NOT_POPT,   Option::CHECK,
          "valkyrie",   '\0',               "use-system-font", 
          "",          "true|false",       "true", 
          "Use default system font",  "",  urlValkyrie::userFont );
  addOpt( FONT_USER,   Option::NOT_POPT,   Option::LEDIT, 
          "valkyrie",  '\0',               "user-font", 
          "",          "",  "Luxi Sans,10,-1,5,50,0,0,0,0,0", 
          "",          "",                 urlValkyrie::userFont );
  addOpt( SRC_EDITOR,  Option::NOT_POPT,   Option::LEDIT, 
          "valkyrie",  '\0',               "src-editor", 
          "",          "",                 "/usr/bin/emacs", 
          "Src Editor:",   "",             urlValkyrie::srcEditor );
  addOpt( SRC_LINES,   Option::NOT_POPT,   Option::SPINBOX, 
          "valkyrie",  '\0',               "src-lines",
          "",          "1|10",             "2", 
          "Extra lines shown above/below the target line:", "",
          urlValkyrie::srcLines );
  /* path to valgrind executable (maybe found by configure) */
  addOpt( VG_EXEC,     Option::NOT_POPT,   Option::LEDIT, 
          "valkyrie",  '\0',               "vg-exec",
          "",          "",                 "",
          "Valgrind:", "",                 urlValkyrie::vgDir );
  /* path to a supp files dir. this may have been found by configure,
     but can be changed later via valkyrie's Option page to point to
     another suppression files dir */
  addOpt( VG_SUPPS_DIR,   Option::NOT_POPT, Option::LEDIT, 
          "valkyrie",  '\0',               "vg-supps-dir",
          "",          "",                 "",
          "Supps.Dir:",  "",               urlValkyrie::suppDir );
  addOpt( BINARY,      Option::NOT_POPT,   Option::LEDIT,
          "valkyrie",  '\0',               "binary", 
          "",          "",                 "", 
          "Binary:",   "",                 urlValkyrie::binary );
  addOpt( BIN_FLAGS,   Option::NOT_POPT,   Option::LEDIT,
          "valkyrie",  '\0',               "binary-flags", 
          "",          "",                 "", 
          "Binary flags:", "",             urlValkyrie::binFlags );
  addOpt( VIEW_LOG,    Option::ARG_STRING, Option::NONE, 
          "valkyrie",  '\0',               "view-log", 
          "<file>",    "",                 "",
          "View logfile:", "parse and view a valgrind logfile",
          urlNone );
  /* path to vk_logmerge executable (found by configure) */
  addOpt( MERGE_EXEC,  Option::NOT_POPT,   Option::LEDIT,
          "valkyrie",  '\0',               "merge-exec",
          "",          "",                 "",
          "Log Merger:", "",               urlNone );
  addOpt( MERGE_LOGS,  Option::ARG_STRING, Option::NONE, 
          "valkyrie",  '\0',               "merge", 
          "<loglist>", "",                 "",
          "View logfiles:", "merge multiple logfiles, discarding duplicate errors",
          urlNone );
}


int Valkyrie::checkOptArg( int optid, const char* argval, 
                           bool use_gui/*=false*/ )
{ 
  int errval = PARSED_OK;
  QString argVal( argval );
  Option* opt = findOption( optid );
  switch ( optid ) {

  /* these options are _only_ set via the gui, and are either (a)
     limited to a set of available values, or (b) have already been
     checked, so no need to re-check them. */
    case TOOLTIP:
    case PALETTE:
    case ICONTXT:
    case FONT_SYSTEM:
    case FONT_USER:
    case SRC_LINES:
    case VG_SUPPS_DIR:
      return errval;
      break;

    case SRC_EDITOR:
    case VG_EXEC:
      argVal = binaryCheck( &errval, argval );
      break;

    case VIEW_LOG:
      argVal = fileCheck( &errval, argval, true, false );
      if ( errval == PARSED_OK ) {
        /* check the file format is xml */
        bool ok = XMLParser::xmlFormatCheck( &errval, argVal );
        if ( ok && errval == PARSED_OK ) {
          runMode = modeParseLog;
        }
      }
      break;

    /* expects a single file which contains a list of
       logfiles-to-be-merged, each on a separate line, with a minimum
       of two logfiles. Validating each file is done at merge-time, as
       we will then skip any files which we can't read. */
    case MERGE_LOGS:
      argVal = fileCheck( &errval, argval, true, false );
      if ( errval == PARSED_OK )
        runMode = modeMergeLogs;
      break;

    case BINARY:
      argVal = binaryCheck( &errval, argval );
      if ( errval == PARSED_OK ) 
        runMode = modeParseOutput;
      break;

    case BIN_FLAGS:
      argVal = argval;
      break;
  }

  /* if this option has been called from the cmd-line, save its value
     in valkyrierc if it has passed the checks. */
  if ( errval == PARSED_OK && use_gui == false ) {
    writeOptionToConfig( opt, argVal );
  }

  return errval; 
}


/* get ptrs to all tools. */
void Valkyrie::init()
{
  valgrind   = (Valgrind*)vkConfig->vkObject( "valgrind" );

  ToolList toolList = vkConfig->toolList();

  for ( ToolObject* tool=toolList.first(); tool; tool=toolList.next() ) {
    connect( tool, SIGNAL( fatal() ), this, SLOT( quit() ) );
  }
}


/* slot: called by tools, if they have a fatal() accident */
void Valkyrie::quit()
{ exit(0); }

/* called from MainWindow::run() */
void Valkyrie::setRunMode( Valkyrie::RunMode rm )
{ runMode = rm; }


/* called from Valkyrie::currentFlags() 
   see if valkyrie was told what to do:
   - on the cmd-line
   - via a tool
   - via the gui options page */
QStringList Valkyrie::modifiedFlags()
{
  Option* opt;
  QStringList modFlags;
  QString cfgVal;

  switch ( runMode ) {

    case modeParseLog: {
      opt    = findOption( VIEW_LOG );
      cfgVal = vkConfig->rdEntry( opt->longFlag, name() );
      if ( cfgVal != opt->defaultValue )
        modFlags << "--" + opt->longFlag + "=" + cfgVal;
    } break;

    case modeNotSet:
    case modeParseOutput: {
      opt    = findOption( BINARY );
      cfgVal = vkConfig->rdEntry( opt->longFlag, name() );
      if ( cfgVal != opt->defaultValue ) {
        modFlags << cfgVal;
        /* see if there were any flags given for the binary */
        opt = findOption( BIN_FLAGS );
        cfgVal = vkConfig->rdEntry( opt->longFlag, name() );
        modFlags += QStringList::split(" ", cfgVal);
      }
    } break;

    case modeMergeLogs:
      opt    = findOption( MERGE_LOGS );
      cfgVal = vkConfig->rdEntry( opt->longFlag, name() );
      if ( cfgVal != opt->defaultValue )
        modFlags << "--" + opt->longFlag + "=" + cfgVal;
      break;

  default:
    vk_assert_never_reached();
  }

  return modFlags;
}


/* called by:
   (a) MainWindow::showFlagsWidget() so user can see 
       exactly what is being fed to valgrind on the cmd-line;
   (b) by Valkyrie::runTool().  
   returns a '\n' separated list of current relevant flags */
QString Valkyrie::currentFlags( ToolObject* tool_obj )
{
  flags.clear();

  switch ( runMode ) {

    /* if nothing was specified on the cmd-line, or we're about to run
       valgrind, return the flags which would be used if the current
       tool were run */
    case modeNotSet:
    case modeParseOutput:
      /* get the /path/to/valgrind */
      flags << vkConfig->rdEntry( "vg-exec","valkyrie");
       /* set the tool we are using */
#if 1
      flags << "--tool=" + tool_obj->name();
#else
      flags << "--tool=none";
#endif
      /* check if any valgrind core opts have been modified.
         'modified' means 'set to anything other than default' */
      flags += valgrind->modifiedFlags( tool_obj );
      /* now get flags which have been specified / modified for this
         tool _only_ */
      flags += tool_obj->modifiedFlags();
      /* finally, check for valkyrie-specific flags */
      flags += modifiedFlags();
      break;

    case modeParseLog:
    case modeMergeLogs:
      flags += modifiedFlags();
      break;

  default:
    vk_assert_never_reached();
  }

  /* ### hack alert: unfortunately, we have to pass each arg to
     VKProcess as a separate string, and this includes any binary
     flags; but for display purposes in the flagWidget, concat the
     binary together with its flags 'cos it's prettier. */
  QString flagsStr = flags.join("\n");
  if ( !vkConfig->rdEntry("binary-flags", "valkyrie").isEmpty() )
    flagsStr.replace( flagsStr.findRev('\n'), 1, ' ');
  return flagsStr;
}


/* called from MainWin when user clicks stopButton */
void Valkyrie::stopTool( ToolObject* activeTool )
{
  vk_assert( activeTool != 0 );

  bool success = activeTool->stop( runMode );

  vk_assert( success );  // TODO: what to do if couldn't stop?
}


/* Run the tool for this runMode */
bool Valkyrie::runTool( ToolObject* activeTool )
{
  vk_assert( activeTool != 0 );

  bool success = true;

  /* find out what we are supposed to be doing */
  switch ( runMode ) {

  case modeNotSet:       /* no flags given on cmd-line */
    break;
    
  /* run valgrind --tool=tool_name, with all flags */
  case modeParseOutput:
    currentFlags( activeTool ); /* ensure flags up-to-date */
    success = activeTool->run( flags );
    break;

  /* let tool decide what to do for other runModes */
  case modeParseLog:
  case modeMergeLogs:
    success = activeTool->start( runMode );
    break;

  default:
    vk_assert_never_reached();
  }

  /* we've done what we were asked to do, so relax */
  //runMode = modeNotSet;

  return success;
}


